unit Basic;  interface
uses  Forms, SysUtils, Math, Dialogs, Controls, Buttons, StdCtrls, Classes;
const  cr= #10;  tab= #9;  bigReal= 9e9;  eps=1e-6;
  
type
TDebugList= class(TForm)
  BitBtn: TBitBtn;
  Memo: TMemo;
  procedure FormResize(Sender: TObject);
  procedure BitBtnClick(Sender: TObject);
  procedure FormClose(Sender: TObject; var Action: TCloseAction);
private
  f: TextFile;  FileOpen: boolean;
public
  procedure Add(const s: string);
end;

TVector= class
private
  _X, _Y, _R{value}, _Fi{angle, 0-right, pi/2-down}: real;
  procedure WriteX(const Value: real);
  procedure WriteY(const Value: real);
  procedure WriteR(const Value: real);
  procedure WriteFi(const Value: real);
  procedure SetPolar;
  procedure SetDescart;
public
  procedure SetXY(X,Y: real);
  procedure Add(V: TVector);
  property X: real read _X write WriteX;
  property Y: real read _Y write WriteY;
  property R: real read _R write WriteR;
  property Fi: real read _Fi write WriteFi;
end;

TISet=class {Sorted Array Of Integer}
private
  M: array of integer;
  function ReadElement(index: integer): integer;
public
  function maxIndex: integer;
  function View: string;
  function GE(I: TISet): boolean;
  function EQ(IntSet: TISet): boolean;
  procedure AssignTISet(IntSet: TISet);
  procedure Include(e: integer);
  procedure Exclude(e: integer);
  function Contains(e: integer): boolean;
  function Empty: boolean;
  property Element[index: integer]: integer Read ReadElement; default;
end;

function ro(x: Real): string;  { Float to String }
function ro2(x: Real;  Width: byte): string;
function rnd(b1, b2: integer): integer;
function rnd1: real;{-1..+1}
function P1to(i: integer): boolean; {true with probability P=1/i}
function DirectionTo(x{ right }, y{ down }: Real): Real;  { dir. to point x, y }
function NormPi(Angle: Real): Real;  { ranges angle to - pi...+pi }

function  RunTime: Real; {seconds}

var  DebugList: TDebugList;  Debuglevel: byte= 3;

implementation
{$R *.DFM}

const  txtOk= 'Ok';  txtStop= 'Stop';  txtContinue= 'Continue';
var  StartTime: TDateTime= -1;

function NormPi(Angle: Real): Real;  { ranges angle to - pi...+pi }
begin
  while Angle < -pi do Angle:= Angle+2*pi;
  while Angle > +pi do Angle:= Angle-2*pi;
  NormPi:= Angle;{-pi..+pi}
end;

function  RunTime: Real; {seconds}
const day=24*3600;
begin
  if  StartTime<0  then  StartTime:= Time;
  RunTime:= (Time-StartTime)*day;
end;

procedure TDebugList.BitBtnClick(Sender: TObject);
begin
  if  BitBtn.Caption= txtOk  then  Visible:= false  else
  if  BitBtn.Caption= txtContinue  then  BitBtn.Caption:= txtStop  else
  if  BitBtn.Caption= txtStop  then  BitBtn.Caption:= txtContinue;
end;

procedure TDebugList.Add(const s: string);
var  st: string;  i: integer;
begin
  if not Visible or (BitBtn.Caption= txtContinue) then exit;
  BitBtn.Caption:= txtStop;
  if  not  FileOpen  then
  begin
    st:= 'synt';
    for  i:= 1  to  4  do  st:= st+ char(rnd(ord('a'),ord('z')));
    st:= st+'.txt';
    AssignFile(f, st);  Rewrite(f);  FileOpen:= true;
    st:= 'SYNT log file: '+st+cr+DateToStr(Date)+' '+TimeToStr(Time)+cr;
    writeln(f, st);
  end;
  writeln( f, s);
  if Memo.Lines.Add(s)=0  then;
end;

procedure TDebugList.FormResize(Sender: TObject);
begin  BitBtn.SetBounds(ClientWidth-BitBtn.Width, ClientHeight-BitBtn.Height,
                        BitBtn.Width, BitBtn.Height);
end;

function rnd(b1, b2: integer): integer;
begin
  if  b2>b1  then  rnd:= b1+random(b2-b1+1)
             else  rnd:= b2+random(b1-b2+1);
end;

function ro(x: Real): string;  begin  ro:=Format('%g',[x]);  end;

function ro2(x: Real;  Width: byte): string;
begin
  ro2:= Format('%-*.*g',[Width, 2, x]);
end;

procedure TDebugList.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if  FileOpen  then  begin  CloseFile(f);  FileOpen:=false;  end;
end;

procedure TISet.AssignTISet(IntSet: TISet);
begin
  SetLength(M, Length(IntSet.M));
  move(IntSet.M[0], M[0], Length(M)*SizeOf(M[0]));
end;

function TISet.EQ(IntSet: TISet): boolean;  var  i: integer;  ok: boolean;
begin
  ok:= Length(M)=Length(IntSet.M);
  if  ok  then  for  i:=0  to  High(M)  do
                if  M[i]<>IntSet.M[i]  then  begin  ok:= false;  break;  end;
  EQ:= ok;
end;

procedure TISet.Exclude(e: integer);  var  i, j: integer;
begin  j:= High(M);
  for  i:=0  to  j  do  if  M[i]=e  then
  begin
    if  j>i  then  move(M[i+1], M[i], (j-i)*SizeOf(M[0]));
    SetLength(M, j);  break;
  end;
end;

procedure TISet.Include(e: integer);  var  i, j: integer;
{}procedure  insert;
  begin  j:= Length(M);  SetLength(M, j+1);
    if  j>i  then  move(M[i], M[i+1], (j-i)*SizeOf(M[0]));
    M[i]:= e;
  end;
begin
  i:= 0;
  while  i<=High(M)  do  begin  if  M[i]>=e  then  break;  inc(i);  end;
  if  i>High(M)  then  insert  else  if  M[i]>e  then  insert;
end;

function TISet.ReadElement(index: integer): integer;
begin
  ReadElement:= M[index];
end;

function TISet.View: string;
var  st: string;  i: integer;
begin  st:= '';
  for  i:= 0  to  High(M)  do  st:= st+ro(M[i])+':';
  View:= st;
end;

function TISet.Empty: boolean;  begin  Empty:= Length(M)=0;  end;
function TISet.maxIndex: integer;  begin  maxIndex:= High(M);  end;
{procedure TDebugList.Run;  begin  BitBtn.Caption:= txtStop;  end;}

function TISet.GE(I: TISet): boolean;
var  j: integer;
begin
  GE:= false;
  for  j:= 0  to  High(I.M)  do  if  not Contains(I[j])  then  exit;
  GE:= true;
end;

function TISet.Contains(e: integer): boolean;
var  i: integer;
begin
  Contains:= true;
  for  i:=0  to  High(M)  do  if  M[i]=e  then  exit;
  Contains:= false;
end;

function  rnd1: real;{-1..+1}  begin  rnd1:= (2*random-1);  end;
function  P1to(i: integer): boolean; begin P1to:=rnd(1,i)=i; end;

function DirectionTo(x{ right }, y{ down }: Real): Real;  { dir. to point x, y }
var  a, d: Real;
begin
  d:= sqrt(sqr(x)+sqr(y));  { distance }
  if  d=0  then  a:= rnd1*pi{-pi..+pi}
           else  a:= Math.ArcCos(x/d);{0..+pi}
  if  y<0  then  a:= - a;{- pi..+pi}
  DirectionTo:= a;  { 0 - right, pi/2 - down, +-pi - left, -pi/2 - up }
end;

{ TVector }

procedure TVector.Add(V: TVector);
begin
  if  V.R=0  then  exit;
  _X:=_X+V._X;
  _Y:=_Y+V._Y;
  SetPolar;
end;

procedure TVector.SetDescart;
begin
  _X:=_R*cos(_Fi);
  _Y:=_R*sin(_Fi);
end;

procedure TVector.SetPolar;
begin
  _R:= sqrt(sqr(_X)+sqr(_Y));
  _Fi:= DirectionTo(_X, _Y);
end;

procedure TVector.SetXY(X, Y: real);
begin
  _X:=X;
  _Y:=Y;
  SetPolar;
end;

procedure TVector.WriteFi(const Value: real);
begin
  _Fi:= Value;
  SetDescart;
end;

procedure TVector.WriteR(const Value: real);
begin
  _R:= Value;
  SetDescart;
end;

procedure TVector.WriteX(const Value: real);
begin
  _X:= Value;
  SetPolar;
end;

procedure TVector.WriteY(const Value: real);
begin
  _Y:= Value;
  SetPolar;
end;

end.
