(* Project SYNT demonstrates an algorithm of universal learning.
   The code was tested in DELPHI 7.

   (c) E.Kornienko, 1997-2003
      http://webcenter.ru/~korn/

   You may change and distribute the code.
   In case of distribution you should enclose
     the copy of initial copyrighted code, too.
*)
unit Brain;  interface
uses  Basic;
const  detectEffectors= true;

type
Tsense=-1..254;{maximum diversity of data. Normally, less then 3..7}

TEffect= record{statistics for a variant of effector's action}
  N: integer;
  Sum: Real;{average S is Sum/N}
End;

TProcess= class{the set of actions for a process}
private
  E: array of TEffect;
  _Command: integer;  _DS: Real;
  procedure div2(maxN: integer); {if N>maxN then N/2, sum/2}
  procedure RegisterProcess(i: integer;  Success, midDS: Real);
  function  View: string;
  procedure DebugView;
  property  DS: real  Read _DS;
  property  Command: integer  Read _Command;
end;

TEngram= class(TISet){list of organs, which should work together}
private
  P: array of TProcess{of Effect};
  _DS, _midDS: real;  _Command, _Ts, _History: integer;
  procedure CheckIndex(ip: integer);
public
  Time, waitTime: integer;
  constructor Create(Sample: TEngram);  destructor Destroy;  override;
  procedure Modify(NO: integer);
  procedure debugView(level: integer);
  procedure SetCommand(ip: integer);
  procedure RegisterProcess(ip, Command: integer; Success: real);
  function  Empty: boolean;
  property  DS: real read _DS;
  property  Command: integer read _Command;
  property  midDS: real read _midDS;
  property  Ts: integer read _Ts;
  property  History: integer read _History;
end;

TOrgan= class
private
  E: array of TEngram{of Process of Effect};
  NeedOrgans: array of TOrgan;
  Inputs, Outputs: array of TSense;  {lists of indices of values}
  V: array of Real;  { assortment of data }
  NPredictedChange, NTrueChange, NO: integer;
  Input, Output{ current indices }: TSense;
  EffectLevel, maxW: real;  Effector: boolean;  recentEngram: TEngram;
  procedure Synthesis;
  procedure EngramList(Tc: integer);
  procedure debugView(level: integer);
  procedure NewInput(Tc: integer);
  function  ReadValue: Real;
  procedure WriteValue(Value: Real);
public
  property  Value: Real  read ReadValue  write WriteValue;{in TMoths}
  procedure Demands(Organ: TOrgan);
  destructor Destroy;  override;
end;

TBrain= class{ a neural node}
private
  O: array of TOrgan{of Engram of Process of Effect};
  P: array of single;
  Tc: integer;{current time}  _Pleasure: Real;
  procedure  ScanningMemory(En: TEngram; Org: TOrgan);
  procedure incTc;
public
  Name: string;
  property  Pleasure: Real  write _Pleasure;
  procedure Action(level: integer);
  function  newOrgan: TOrgan;  { used in TMoths }
  destructor Destroy;  override;
end;

implementation
const  maxN= 10000;

procedure  TOrgan.debugView(level: integer);  var  i: integer;
begin
  if  Effector and (level>0) and (Length(E)>0)  then
  for  i:= 0  to  High(E)  do  E[i].debugView(level-1);
end;

procedure  TOrgan.NewInput(Tc: integer);
begin
  Inputs[Tc]:= Input;  Outputs[Tc]:= Output;
  if  Tc>=1  then  if  Outputs[Tc]<>Outputs[Tc-1]  then
  begin  inc(NPredictedChange);{if a predicted event occurs hence the prediction
                                     is a command and the organ is an effector }
    if  Inputs[Tc]= Outputs[Tc]  then
    if  Inputs[Tc-1]= Outputs[Tc-1]  then  inc(NTrueChange);
    EffectLevel:= NTrueChange/NPredictedChange*Length(V);
    if  detectEffectors  then  Effector:= EffectLevel>=0.5
                         else  Effector:= true;{  -  }
    if  NPredictedChange>maxN {    }  then
    begin
      NPredictedChange:= NPredictedChange div 2;
      NTrueChange:= NTrueChange div 2;
    end;
  end;
end;

destructor TOrgan.Destroy;  var  i: integer;
begin  for  i:= 0  to  High(E)  do  E[i].Destroy;  inherited;  end;

function  TOrgan.ReadValue: Real;
begin
  if  (Output>=0) and (Output<=High(V))
  then  ReadValue:= V[Output]  else  ReadValue:= 1000*rnd1;{-1000..+1000}
end;

procedure TOrgan.WriteValue(Value: Real);  var  i: integer;
begin
  for  i:= 0  to  High(V)  do
  if  abs(Value-V[i])<eps  then  begin  Input:= i;  exit;  end;
  if  High(V)>=High(Input)  then  exit;
  SetLength(V, Length(V)+1);  V[High(V)]:= Value;  Input:= High(V);
end;

procedure TBrain.Action(level: integer);
var  io, ie, i, j: integer;  st: string;
begin
  incTc;  P[Tc]:= _Pleasure;
  for  io:= 0  to  High(O)  do  O[io].newInput(Tc);
  for  io:= 0  to  High(O)  do  with  O[io]  do{separate cycle}
  begin  for  ie:= 0  to  High(E)  do  ScanningMemory(E[ie], O[io]);
    NO:= Length(O);
    EngramList(Tc);
      if  High(E)>=0  then
      for  j:= 0  to  High(O)  do  if  not E[High(E)].Contains(j) then
      for  i:= 0  to  High(NeedOrgans)  do
      if  O[j]=NeedOrgans[i]  then  E[High(E)].Include(j);
    Synthesis;
  end;
{listing}
  st:='[Pl='+ro2(P[Tc],2)+']';
  if  (Level>0) and (Length(O)>0)  then
    DebugList.Add(Name+tab+' Tc='+ro(Tc)+'  Pl='+ro2(_Pleasure,4)
                  +' NO='+ro(Length(O))+st);
  for  io:= 0  to  High(O)  do
  begin
    if  level>1  then  with  O[io]  do  if  Effector  then
      DebugList.Add(' Organ'+ro(io)+' NE='+ro(Length(E))
        +' NV='+ro(Length(V))+' Efl='+ro2(EffectLevel, 4));
    O[io].debugView(level-1);
  end;
end;

function TBrain.newOrgan: TOrgan;
begin
  SetLength(O, Length(O)+1);  O[High(O)]:= TOrgan.Create; newOrgan:= O[High(O)];
end;

destructor TBrain.Destroy;  var  i: integer;
begin  for  i:=0  to  High(O)  do  O[i].Destroy;  inherited;  end;

procedure TBrain.incTc;  var  ie, io, t1, len: integer;
begin
  len:= length(P);
  if len>=maxN then
  begin
    t1:= len div 2;  len:= len-t1;
    move(P[t1], P[0], len*SizeOf(P[0]));  SetLength(P, len);
    for  io:= 0  to  High(O)  do  with  O[io]  do
    begin
      move(Inputs[t1], Inputs[0], len*SizeOf(Inputs[0]));
      move(Outputs[t1], Outputs[0], len*SizeOf(Outputs[0]));
      SetLength(Inputs, len);  SetLength(Outputs, len);
      for  ie:= 0  to  High(E)  do  with  E[ie]  do
      if  Time<t1  then  Time:= 0  else  Time:= Time-t1;
    end;
  end;
  len:= Length(P)+1;  SetLength(P, len);
  for  io:= 0  to  High(O)  do  with  O[io]  do
  begin  SetLength(Inputs, len);  SetLength(Outputs, len);  end;
  Tc:= High(P);
end;

procedure  TOrgan.Synthesis;
var  i, imax, Com: integer;  Eb: TEngram;
begin
  Com:= -1;
  if  High(E)>=0  then  if  Effector  then
  begin
    maxW:= E[0].midDS;  imax:=0;  Eb:= nil;
    for  i:= 0  to  High(E)  do
    begin
      if  E[i]=recentEngram  then  Eb:= recentEngram;
      if  E[i].midDS>maxW  then  begin  maxW:= E[i].midDS;  imax:= i;  end;
    end;
    for  i:= 0  to  High(E)  do
    begin
      if  Eb=nil  then  Eb:= E[rnd(0,High(E))];
      if  (E[i].DS>2*Eb.DS) and (E[i].midDS>maxW/2)  then  Eb:= E[i];
      if  Eb.midDS<maxW/4  then  Eb:= E[imax];
    end;
    RecentEngram:= Eb;  Com:= Eb.Command;
    for  i:= 0  to  High(E)  do  with  E[i]  do
    if  (Command=Com) and (Com>=0)  then  WaitTime:=0
    else  if  E[i]<>Eb  then  inc(WaitTime);
  end;
  if  (Com>=0) and (Com<=High(V))  then  Output:= Com
                                   else  Output:= rnd(-1, High(V));{need!}
end;

procedure TOrgan.EngramList(Tc: integer);
{}procedure delEngram(i: integer);
  begin  E[i].Destroy;
    if  i<High(E)  then  move(E[i+1], E[i], (High(E)-i)*SizeOf(E[0]));
    SetLength(E, Length(E)-1);
  end;
const  maxWaitTime= 8;  minTime=100;
var  maxE, i: integer;  En: TEngram;
begin
  if  (not Effector) or (Tc<maxWaitTime)  then  exit;
  maxE:= NO;  if  maxE>6  then  maxE:= 6;
  i:= 0;
  while  i<=High(E)  do  with  E[i]  do
  if  (Time>minTime)
   and ((WaitTime>maxWaitTime) and ((midDS<maxW/2) and (Length(E)>0.25*maxE))
    or ((WaitTime>4*maxWaitTime) and (midDS<0.9*maxW)))
  then  delEngram(i)  else  inc(i);
  if  High(E)>=0  then  if  E[High(E)].Time<Tc div maxE  then  exit;
  if  Length(E)>= maxE  then  exit;
  En:= nil;
  if  High(E)>=0  then  if  P1to(2)  then  En:= E[rnd(0, High(E))];
  SetLength(E, Length(E)+1);  E[High(E)]:= TEngram.Create(En);  En:= E[High(E)];
  En.Modify(NO);
  if  En.Empty  then  delEngram(High(E));
end;

procedure  TBrain.ScanningMemory(En: TEngram; Org: TOrgan);
{}function Command(t: integer): integer;  var  c: integer;
  begin  c:= Org.Outputs[t];
    if  (c>=0) and (c<=High(Org.V))  then  Command:= c
                                     else  Command:= Org.Inputs[t];
  end;
{}function  IndexOfProcess(t: integer): integer;  var  ip, io, h: integer;
  begin  ip:= 0;
    for  io:= 0  to  En.maxIndex  do  for  h:= 0  to  En.History-1  do
    with  O[En[io]]  do  ip:= ip*Length(V)+Inputs[t-h];
    IndexOfProcess:= ip;
  end;
{}function Success(t: integer): real;  var  i, n: integer;  x: real;
  begin  x:= 0;  n:= 0;
    for  i:= t-1  to  t+En.Ts  do
    begin inc(n);
      if (i>0) and (i<Tc) then  if  i<t  then  x:= x-P[i]  else  x:= x+P[i];
    end;
    if  n<>0  then  x:= x/n;  Success:= x;
  end;
var  i, t1: integer;
begin
  if  Org.Effector  then
  begin
    for  i:= 1  to  2  do  if  En.Time<Tc-En.Ts-1  then
    begin  t1:= En.Time+1;
      if  En.Time>=En.History  then
        En.RegisterProcess(IndexOfProcess(En.Time), Command(t1), Success(t1));
      En.Time:= t1;
    end;
    En.SetCommand(IndexOfProcess(Tc));
  end;
end;

procedure TOrgan.Demands(Organ: TOrgan);
var  i: integer;
begin
  for  i:= 0  to  High(NeedOrgans)  do  if  NeedOrgans[i]=Organ  then  exit;
  SetLength(NeedOrgans, Length(NeedOrgans)+1);
  NeedOrgans[High(NeedOrgans)]:= Organ;
end;

procedure TProcess.div2(maxN: integer);
var i: integer;
begin
  for  i:= 0 to High(E)  do  with  E[i]  do
  if  N>maxN  then  begin  sum:= sum*(N div 2)/N;  N:= N div 2; end;
end;

procedure TProcess.RegisterProcess(i: integer;  Success, midDS: Real);
const dt=20;  minN=10;  maxN=1000;
var  Si, Smin, Smax: Real;  Imax: integer;
begin
  while  i>High(E)  do
  begin  SetLength(E, Length(E)+1);  E[High(E)].N:= 0;  E[High(E)].sum:= 0; end;
  with  E[i]  do
  begin
    inc(N);  sum:= sum+success;  if  N>maxN  then  div2(maxN);  Smin:= Sum/N;
  end;
  Smax:= Smin;  Imax:= i;
  for  i:= 0  to  High(E)  do  with  E[i]  do  if  N<>0  then
  begin  Si:= Sum/N;
     if  Si>Smax  then  begin  Smax:= Si;  Imax:= i;  end;
     if  Si<Smin  then  Smin:= Si;
  end;
  _DS:= Smax-Smin;
  if  DS<midDS/2  then  div2(minN);{DS is too small for correct choice}
  if  P1to(10*dt)  then  _Command:= rnd(-1, High(E)){creative random choice};
  if  P1to(dt)    then  _Command:= Imax{best choice};
  if  DS<=0       then  _Command:= -1{initial random choice};
end;

function TProcess.View: string;
begin
  View:=' com='+ro(Command)+' NE='+ro(Length(E))+' DS='+ro2(DS, 6);
end;

procedure TProcess.DebugView;
var  i: integer;
begin
  for  i:= 0  to  High(E)  do  with  E[i]  do  if  N<>0  then
    DebugList.Add('E'+ro(i)+tab+'S='+ro2(sum/N, 4)+tab+' N='+ro(N));
end;

constructor TEngram.Create(Sample: TEngram);
begin
  inherited Create;  _History:= 1;  _Ts:= 20;
  if  Sample<>nil  then
  begin
    AssignTISet(Sample);  _History:= Sample.History;  _Ts:= Sample._Ts;
  end;
end;

procedure TEngram.debugView(level: integer);
var  i: integer;  st: string;
begin
  st:= View;
  while  Length(st)<8  do  st:=' '+st;
  if  level>0  then  DebugList.Add('E'+st+'H'+ro(History)+' Ts='+ro(Ts)
    +tab+'  NP='+ro(Length(P))
    +'  t='+ro(time)+'  Wt='+ro(WaitTime)+'  (DS='+ro2(DS,4)+')'
    +' mDS='+ro2(midDS,2));
  for  i:= 0  to  High(P)  do
  begin
    if  level>1  then  DebugList.Add('Pr'+ro(i)+tab+P[i].View);
    if  level>2  then  P[i].DebugView;
  end;
end;

destructor  TEngram.Destroy;
var  i: integer;
begin
  for  i:= 0  to  High(P)  do  P[i].Destroy;  inherited;
end;

procedure TEngram.CheckIndex(ip: integer);
begin
  while  ip>High(P)  do
  begin  SetLength(P, Length(P)+1);  P[High(P)]:= TProcess.Create;  end;
end;

procedure TEngram.SetCommand(ip: integer);
const a=0.01;
begin
  CheckIndex(ip);
  _Ds:= P[ip].DS;
  _Command:= P[ip].Command;
  _midDS:= (1-a)*midDS+a*DS;
end;

procedure TEngram.RegisterProcess(ip, Command: integer; Success: real);
begin
  CheckIndex(ip);
  P[ip].RegisterProcess(Command, Success, midDS);
end;

procedure TEngram.Modify(NO: integer);
const  minTs=2;  maxTs=64;
begin
  if  P1to(100)  then  dec(_History)  else  Exclude(rnd(0, NO-1));
  if  P1to(100)  then  inc(_History)  else  Include(rnd(0, NO-1));
  if  P1to(2)   then  _Ts:= Ts div 2  else  _Ts:= Ts * 2;
  if  Ts<minTs  then  _Ts:= minTs  else  if  Ts>maxTs  then  _Ts:= maxTs;
end;

function TEngram.Empty: boolean;
begin
  Empty:= inherited Empty or (History<=0);
end;

end.
