(* Project SYNT demonstrates an algorithm of universal learning.
   The code was tested for DELPHI 7.

   (c) E.Kornienko, 1997-2003
      http://webcenter.ru/~korn/

   You may change and distribute the code.
   In case of distribution you should enclose
     the copy of initial copyrighted code, too.

     ,   TImage,  
    .     .
  ,     ,   .
*)
unit Moths;  interface

uses  Basic, Brain,
  Graphics,  Classes, ExtCtrls, Controls, Windows, Forms;

var  AllowMotion: boolean= true;  { menu switcher }

type

{      }
TBallClass= class of TBall;
TBall= class(ExtCtrls.TImage)
  constructor Create(AOwner: Classes.TComponent);  override;
  destructor Destroy; override;
  procedure MouseDown
    (B: Controls.TMouseButton;  S: TShiftState;  X, Y: integer);  override;
  procedure MouseMove(S: TShiftState;  X, Y: integer);  override;
  procedure MouseUp(B: TMouseButton;  S: TShiftState;  X, Y: integer);  override;
private
  solid: boolean;
  V{velocity, pixel/move}, A{drive acceleration, 0..1},
  F{face view, body oientation}: TVector;
  Vmax{pixel/move},
  Ta{acceleration to Vmax, moves}, Td{decceleraion to 0, moves}: real;
  MouseGrab: boolean;  Xgrab, Ygrab: integer;
  function  TouchBy(image: TImage): boolean;
  function  FindNear(body: TBallClass): TImage;
  function  DirTo(image: TImage): Real;{,    image}
  procedure SetHint;
  procedure rndJump;
public
  procedure Action;  virtual;{ }
end;

{  .      .}
TTarget= class(TBall)
  constructor Create(AOwner: TComponent);  override;
private
  hIcon: Windows.hIcon;{ }
public
  procedure Action; override;
end;

{   ,       }
TRose=class(TBall)
  constructor Create(AOwner: TComponent);  override;
private
  hSmall, hBig, hNormal: Windows.hIcon;
  t0: real;{  }
public
  procedure Action; override;
end;

{ TBall }
TSensing= class(TBall)
  constructor Create(AOwner: TComponent);  override;
  destructor Destroy; override;
private
  Pleasure: real;
  Brain: TBrain;
  procedure CreateOrgans(Brain: TBrain); virtual; abstract;
  procedure ChangeBrainBy(otherBrain: TBrain);
end;

{,    TBall,     
  }
TEye= class(TSensing)
  constructor Create(AOwner: TComponent);  override;
  destructor  Destroy; override;
private
{     }
  ib: array[0..15] of Graphics.TBitMap;
  Touch: TOrgan;{ }
  BallVision: TOrgan;  {left|right to current Direction of view }
  Drive: TOrgan;  { Turn left|right to current Direction of body orientation }
  procedure MakeBitmaps;
protected
  procedure BallVisionSensation;
  procedure TurnAction;
  procedure CreateOrgans(Brain: TBrain); override;
public
  procedure Action;  override;
end;

{  }
TButterfly= class(TSensing)
  constructor Create(AOwner: TComponent);  override;
private
  hIcon: Windows.hIcon;
  Touch: TOrgan;
  RoseVision: TOrgan;  {SideVision left|right to current D }
  Drive: TOrgan;  { Turn left|right to current D of motion }
protected
  procedure RoseVisionSensation;
  procedure TurnAction;
  procedure CreateOrgans(Brain: TBrain); override;
public
  procedure Action;  override;
end;

{       .
     }
TMultiBody= Class(TBall)
  destructor Destroy; override;
  constructor Create(AOwner: TComponent);  override;
private
  CommonBrain: TBrain;
  B: array of TBall;
public
  procedure Action;  override;
  procedure newBody(Body: TBallClass);
end;

procedure Register;

implementation
const  maxV=8;

procedure Register;  { Placing the components onto the Delphi's Pallete }
begin
  RegisterComponents('Moths', [TButterFly, TRose, TTarget, TEYE]);
end;

procedure  TMultiBody.Action;
var  i: integer;  P: Real;
begin
{ ,   }
  P:= 0;
  for  i:= 0  to High(B)  do  if  B[i] is TSensing  then
  with  B[i] as TSensing  do  begin  P:= P+Pleasure;  Pleasure:= 0;  end;
{  }
  CommonBrain.Pleasure:= P;
  CommonBrain.Action(DebugLevel);
end;

destructor TMultiBody.Destroy;
var  i: integer;
begin
  for  i:= 0  to High(B)  do  B[i].Destroy;  CommonBrain.Destroy;
  inherited;
end;

constructor TMultiBody.Create(AOwner: TComponent);
begin
  inherited;
  CommonBrain:= TBrain.Create;
  SetHint;  CommonBrain.Name:= Hint;
{  .    }
  Hide;
end;

{  .     .}
procedure TMultiBody.newBody(Body: TBallClass);
var  i: integer;
begin
  i:= Length(B);
  SetLength(B, i+1);
  B[i]:= Body.Create(Owner);
  if  B[i] is TSensing  then
    (B[i] as TSensing).ChangeBrainBy(CommonBrain);
  B[i].Hint:= B[i].Hint+'(at_'+Hint+')';
end;

{ TBall }

procedure TBall.Action;
var  dt: real;  image: TImage;  Vc, dx: TVector;
     ball: TBall;  L,T: integer;  sbl: boolean;
begin
  if  MouseGrab  then  exit;
  dt:= 1;  L:= Left;  T:= Top;
{ interaction with a window's wall }
    if  Left<0  then  begin  L:= 0;  if  V.X<0  then  V.X:= -V.X;  end;
    if  Left+Width>Parent.ClientWidth  then
    begin  L:= Parent.ClientWidth-Width;  if  V.X>0  then  V.X:= -V.X;  end;
    if  Top<0  then   begin  T:= 0;  if  V.Y<0  then  V.Y:= -V.Y;  end;
    if  Top+Height>Parent.ClientHeight then
    begin  T:= Parent.ClientHeight-Height;  if  V.Y>0  then  V.Y:= -V.Y;  end;
    if  (L<>Left) or (T<>Top)  then  SetBounds(L, T, Width, Height);
{ interaction with balls }
  image:= FindNear(TBall);
  if  solid  then
  if  image is TBall  then  if  TouchBy(image)  then
  begin
    ball:= image as TBall;
    Vc:= TVector.Create;
      Vc.SetXY(0.5*(V.X+ball.V.X), 0.5*(V.Y+ball.V.Y));
      V.SetXY(V.X-Vc.X, V.Y-Vc.Y);
      ball.V.SetXY(ball.V.X-Vc.X, ball.V.Y-Vc.Y);
      dx:= TVector.Create;
        dx.SetXY(ball.Left-Left, Ball.Top-Top);
        sbl:= (dx.X*V.X+dx.Y*V.Y)>0;
      dx.Destroy;
      if  sbl  then
      begin
        V.Fi:= V.Fi+pi;
        ball.V.Fi:= ball.V.Fi+pi;
        V.SetXY(V.X+Vc.X, V.Y+Vc.Y);
        ball.V.SetXY(ball.V.X+Vc.X, ball.V.Y+Vc.Y);
      end;
    Vc.Destroy;
  end;
{ mechanical forces }
  V.R:= V.R+(A.R*cos(A.Fi-V.Fi)-V.R/Ta-Vmax/Td)*dt;  if  V.R<0  then  V.R:=0;
  V.Fi:=V.Fi+DirectionTo(V.R, A.R*dt*sin(A.fi-V.fi));
{ reset A.R after it is used }
  A.R:=0;
  SetBounds(round(Left+V.X*dt), round(Top+V.Y*dt), Width, Height); // move done
end;

constructor TBall.Create(AOwner: TComponent);
begin
  inherited;
  V:= TVector.Create;{}
  A:= TVector.Create;{}
  F:= TVector.Create;{  ()}
  Parent:= TWinControl(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  AutoSize:= true;  ShowHint:= true;  Enabled:=true;  Show;  SetHint;
  Transparent:= true;
  Ta:=20;  Td:=200;  Vmax:= 20;
  solid:= {true}false;
end;

destructor TBall.Destroy;
begin
  V.Destroy;  A.Destroy;  F.Destroy;
  inherited;
end;

function TBall.DirTo(image: TImage): Real;
begin
  if  Image<> nil  then  dirTo:= DirectionTo(Image.Left-Left, Image.Top-Top)
                   else  dirTo:= rnd1*pi;
end;

{    }
function TBall.FindNear(body: TBallClass): TImage;
var  i, x, minDistance: integer;  t: TBall;  image: TImage;
begin
  image:= nil;
  minDistance:= maxInt;
  for  i:= 0  to  Owner.ComponentCount-1  do
  if  Owner.Components[i] is body  then
  begin
    t:= Owner.Components[i] as TBall;
    if  t <> Self  then  if  t.Visible  then
    begin
      x:= sqr(Left - t.Left)+sqr(Top - t.Top);
      if x < minDistance then  begin  image:= t;  minDistance:= x;  end;
    end;
  end;
  FindNear:= image;
end;

{  }
procedure TBall.MouseDown(B: TMouseButton; S: TShiftState; X, Y: integer);
begin
  inherited;
  MouseGrab:= true;
  Xgrab:= X;
  Ygrab:= Y;
end;

{    }
procedure TBall.MouseMove(S: TShiftState; X, Y: integer);
begin
  inherited;
  if  MouseGrab  then  SetBounds(Left+X - Xgrab, Top+Y - Ygrab, Width, Height);
end;

procedure TBall.MouseUp(B: TMouseButton; S: TShiftState; X, Y: integer);
begin
  inherited;
  MouseGrab:= false;
end;

{   }
procedure TBall.rndJump;
begin
  SetBounds(rnd(0,Parent.ClientWidth)-Width,
            rnd(0,Parent.ClientHeight)-Height, Width, Height);
end;

{    }
procedure TBall.SetHint;
var  i, c: integer;  st: string;
begin
  c:= 0;
  for  i:= 0  to  Owner.ComponentCount-1  do
  if  Owner.Components[i].ClassName=ClassName  then  inc(c);
  st:= ClassName+ro(c);  delete(st, 1, 1);
  Hint:= st;
end;

{   , }
function TBall.TouchBy(image: TImage): boolean;
begin
  if  image= nil  then  TouchBy:= false  else
  TouchBy:= (Left+Width >= image.Left) and (Left <= image.Left+image.Width) and
            (Top+Height >= image.Top)  and (Top <= image.Top+image.Height);
end;

{ TTarget }

procedure TTarget.Action;
begin
  inherited;
  if  hIcon=0  then
  begin
    hIcon:=  LoadIcon(hInstance, 'BULL');
    Picture.Icon.Handle:= hIcon;
  end;
  Td:= Parent.ClientHeight; Ta:=100*Td;
  if  not AllowMotion  then  V.R:= 0;
  if  AllowMotion and (V.R<Vmax/1e5)  then
  begin
    V.R:= Vmax/2;  V.Fi:= rnd1*pi;
  end;
end;

constructor TTarget.Create(AOwner: TComponent);
begin
  inherited;
  SetHint;
  if csDesigning in ComponentState  then
    Picture.Bitmap.LoadFromResourceName(HInstance, 'TTARGET');
  SendToBack;
  rndJump;
end;

{ TSensingBall }

procedure TSensing.ChangeBrainBy(otherBrain: TBrain);
begin
  Brain.Destroy;
  Brain:= nil;{ ,       }
  CreateOrgans(otherBrain);
end;

constructor TSensing.Create(AOwner: TComponent);
begin
  inherited;
  Brain:= TBrain.Create;
end;

destructor TSensing.Destroy;
begin
  if  Brain<>nil  then  Brain.Destroy;
  inherited;
end;

{ TBallFly }

procedure TEye.Action;
begin
  inherited;{Ball motion}
{ .   } MakeBitMaps;
  if  Brain<>nil  then
  begin
    Brain.Pleasure:= Pleasure;  Brain.Action(DebugLevel-1);
  end;
  Pleasure:= 0;
  TurnAction;{action by effectors}
{sensation by sensors: }
  Touch.Value:= 0;
  if TouchBy(FindNear(TRose))  then
  begin  Touch.Value:= 1;  Pleasure:= Pleasure+10;  end;
  BallVisionSensation;
end;

constructor TEye.Create(AOwner: TComponent);
var  i: integer;
begin
  inherited;
  CreateOrgans(Brain);  SetHint;
  for i:= Low(ib)  to  High(ib)  do  ib[i]:= Graphics.TBitMap.Create;
  if csDesigning in ComponentState  then
    Picture.Bitmap.LoadFromResourceName(HInstance, 'TEYE');
  Brain.Name:= Hint;
  rndJump;
end;

procedure TEye.CreateOrgans(Brain: TBrain);
begin
  Touch:= Brain.newOrgan;
  BallVision:= Brain.newOrgan;
  Drive:= Brain.newOrgan;
  Drive.Demands(BallVision);
{Demands ,       Drive  
  ,   BallVision.   .
  ,   .}
end;

procedure TEye.TurnAction;
const  dFa=pi/8;
var  turn, i: integer;  fi: real;
begin {rate of learning depends on consistency of this algorithm}
  turn:= round(Drive.Value) and 1; // correction of drive.value
  Drive.Value:= turn;
  F.R:=1;
{    dFa}
  if  turn=0  then  F.fi:=F.fi-dFa  else  F.fi:=F.fi+dFa;
{     F}
  A.Add(F);
{  }
  fi:= NormPi(F.Fi)+pi;{0..2pi}
  i:= round((High(ib)-Low(ib)+1)*fi/(2*pi));
  if  i>High(ib)  then  i:= Low(ib);
  Picture.Bitmap:= ib[i];
end;

procedure TEye.BallVisionSensation;
begin
   if  NormPi(DirTo(FindNear(TBall))-A.fi)>0  then  BallVision.Value:=0
                                              else  BallVision.Value:=1;
end;

destructor TEye.Destroy;
var  i: integer;
begin
  for  i:=Low(ib) to High(ib)  do   ib[i].Destroy;
  inherited;
end;

{ TRose }

procedure TRose.Action;
const lifetime=10;
var touch: boolean;  t: real;
begin
  inherited;
  if  hSmall=0  then  hSmall:= LoadIcon(hInstance, 'SMALL_FLOWER');
  if  hNormal=0  then  hNormal:= LoadIcon(hInstance, 'NORMAL_FLOWER');
  if  hBig=0  then  hBig:= LoadIcon(hInstance, 'BIG_FLOWER');
  { checking touch and making simple reaction to the touch }
  V.R:= 0;
  touch:= TouchBy(FindNear(TButterfly));
  t:= RunTime-t0;
  if  (t<1) or (t>(lifetime+3)) then  Picture.Icon.Handle:= 0        else
  if  (t<2) or (t>(lifetime+2)) then  Picture.Icon.Handle:= hSmall   else
  if  (t<3) or (t>(lifetime+1)) then  Picture.Icon.Handle:= hNormal  else
                                      Picture.Icon.Handle:= hBig;
  if touch  then  if  (t>4) and (t<lifetime)  then  t0:=RunTime-lifetime;
  if  t>(lifetime+4)  then  begin  rndJump;  t0:= RunTime;  end;
end;

constructor TRose.Create(AOwner: TComponent);
begin
  inherited;
  SetHint;
  if csDesigning in ComponentState  then
    Picture.Bitmap.LoadFromResourceName(HInstance, 'TROSE');
  SendToBack;
  rndJump;
  solid:= false;
  t0:=RunTime;
end;

{ TButerflyFly }

procedure TButterfly.Action;
begin
  inherited;{Ball motion}
  if  hIcon=0  then
  begin
     hIcon:=LoadIcon(hInstance, 'BLUE_SHADOW_BUTTERFLY');
     picture.icon.Handle:= hIcon;
  end;
  if  Brain<>nil  then
    begin  Brain.Pleasure:= Pleasure;  Brain.Action(DebugLevel-1);  end;
  Pleasure:= 0;
  TurnAction;{action by effectors}
{sensation by sensors: }
  Touch.Value:= 0;
  if TouchBy(FindNear(TRose))  then
  begin  Touch.Value:= 1;  Pleasure:= Pleasure+10;  end;
  RoseVisionSensation;
end;

constructor TButterfly.Create(AOwner: TComponent);
begin
  inherited;
  CreateOrgans(Brain);  SetHint;
  if csDesigning in ComponentState  then
    Picture.Bitmap.LoadFromResourceName(HInstance, 'TBUTTERFLY');
  Brain.Name:= Hint;
  rndJump;
end;

procedure TButterfly.CreateOrgans(Brain: TBrain);
begin
  Touch:= Brain.newOrgan;
  RoseVision:= Brain.newOrgan;
  Drive:= Brain.newOrgan;
  Drive.Demands(RoseVision);
end;

procedure TButterfly.RoseVisionSensation;
begin
   if  NormPi(DirTo(FindNear(TRose))-A.fi)>0  then  RoseVision.Value:=0
                                              else  RoseVision.Value:=1;
end;

procedure TButterfly.TurnAction;
const  dFa=pi/8;
var  turn: integer;
begin {rate of learning depends on consistency of this algorithm}
  turn:= round(Drive.Value) and 1; // correction of drive.value
  Drive.Value:= turn;
  F.R:=1;
  if  turn=0  then  F.fi:=F.fi-dFa  else  F.fi:=F.fi+dFa;
  A.Add(F);
end;

{ .   }
procedure TEye.MakeBitmaps;
const CircleColor=$F0F0F0;  R0=16;
var  i, x, y, Rc, r: integer;  fi: real;
begin
  if  ib[0].Width=0  then
  begin
    picture.Icon.Handle:=LoadIcon(hInstance, 'BLUE_SHADOW_BUTTERFLY');
    for  i:=Low(ib)  to  High(ib)  do  with  ib[i]  do
    begin
      fi:= 2*pi*(i-Low(ib))/(High(ib)-Low(ib)+1);
      Height:= 2*R0;
      Width:= 2*R0;
      Rc:= R0 div 2;
      with  Canvas  do
      begin
        Ellipse(0, 0, Width, Height);{eyeball}
        Brush.Color:= CircleColor;
        FloodFill(Width div 2, Height div 2, Pen.Color, fsBorder);
        x:= Width div 2 - round(Rc*cos(fi));
        y:= Height div 2 - round(Rc*sin(fi));
        r:= Rc;
        Ellipse(x-r, y-r, x+r, y+r);{blue eyepiece}
        Brush.Color:= clBlue;
        FloodFill(x, y, CircleColor, fsSurface);
        r:= r div 2;
        Ellipse(x-r, y-r, x+r, y+r);{center of the eyepiece}
        Brush.Color:= clBlack;
        FloodFill(x, y, clBlue, fsSurface);
      end;
    end;
  end;
end;

end.
