(* Project SYNT demonstrates an algorithm of universal learning.
   The code was tested in DELPHI 7.

   (c) E.Kornienko, 1997-2003
      http://webcenter.ru/~korn/

   You may change and distribute the code.
   In case of distribution you should enclose
     the copy of initial copyrighted code, too.
*)
unit World;  interface
uses  Basic, Moths,
  Dialogs, Forms, Menus, Classes, SysUtils, Graphics, Controls,
  StdCtrls, ExtCtrls;

type
TField= class(TForm)
  MainMenu: TMainMenu;
  Timer: TTimer;
  m_Info: TMenuItem;
  m_About: TMenuItem;
  m_Teaching: TMenuItem;
  m_Consist: TMenuItem;
  m_Know: TMenuItem;
  m_Sensors: TMenuItem;
  m_Flower: TMenuItem;
  m_Motion: TMenuItem;
  m_Control: TMenuItem;
  m_div2: TMenuItem;
  m_div3: TMenuItem;
  m_div4: TMenuItem;
  m_Exit: TMenuItem;
    m_Listing: TMenuItem;
    m_ResetListing: TMenuItem;
  procedure TimerTick(Sender: TObject);
  procedure mAbout(Sender: TObject);
  procedure mTeaching(Sender: TObject);
  procedure mConsist(Sender: TObject);
  procedure mKnow(Sender: TObject);
  procedure mSensors(Sender: TObject);
  procedure mFlower(Sender: TObject);
  procedure mMotion(Sender: TObject);
    procedure mControl(Sender: TObject);
    procedure mExit(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure m_ListingClick(Sender: TObject);
    procedure m_ResetListingClick(Sender: TObject);
private
  TMBody: TMultiBody;
  BRose: TRose;
  Butterfly: TButterfly;
  procedure Action;
  procedure SetDebug(level: integer);
end;

var  Field: TField;

implementation  {$R *.DFM}

const  pre='===   ';  suf='   ==='+cr+cr;

resourcestring
  rsabout=
    '(c) Eugeny Kornienko, 1998-2003'+cr
   +'http: //webcenter.ru/~korn/'+cr+cr
   +'The program demonstrates self-learning of a dynamic associative brain.';
  rsteaching=
    'Make the window small. It is easyer for the Butterfly to learn.'+cr+cr
   +'Place the Flower near the Butterfly. The closer the Flower, the faster '
   +'self-learning.'+cr+cr
   +'Normally it needs about a thousand steps the Butterfly to guess '
   +'how its Pleasure relates to the Flower.';
  rsconsist=
    'The Butterfly has sensors, effectors and a brain.'+cr+cr
   +'Each moment the brain generates behavior that is merely numbers.'+cr+cr
   +'For Effectors these numbers are directions of the Butterfly''s motion.';
  rsknow=
    'After some experience the Butterfly gets to know how and why it may fly '
   +'to the Flower.'+cr+cr
   +'The brain does not know what "motion" '
   +'or "direction" mean. It is only trying to associate its input, output and '
   +'the Butterfly''s Pleasure. The Pleasure is a goal function for the brain.';
  rssensors=
    'Vision.'+cr
   +'Measures the direction to the Flower. It sends a number of the direction '
   +'to the brain.'+cr+cr
   +'Touch.'+cr
   +'Checks touch of the Flower. While touching the Flower it increases '
   +'Pleasure. Beyond the Flower the value of the Pleasure is small.';
  rsflower=
    'The Flower is merely a picture.'+cr+cr
   +'Unlike normal flower it can jump or move to teach the Butterfly.';

procedure TField.TimerTick(Sender: TObject);
var  t: real;  dt0, dt: integer;
begin
  dt0:= Timer.Interval;
  Timer.Interval:= 0; //disabling timer
    t:= RunTime;
    Action;
  dt:= 50*round(1000*(RunTime-t)+1);{ms}
  if  dt>dt0  then  dt:= 2*dt0;
  Timer.Interval:= dt;
  Field.Caption:= ro(round(t))+'s  dt='+ro(dt)+'ms';
  if Debuglevel >= 2 then  DebugList.Add(' ');
end;

procedure TField.mAbout(Sender: TObject);
begin  ShowMessage(pre+'The self-learning Butterfly.'+suf+rsabout);  end;

procedure TField.mTeaching(Sender: TObject);
begin  ShowMessage(pre+m_Teaching.Caption+suf+rsteaching);  end;

procedure TField.mConsist(Sender: TObject);
begin  ShowMessage(pre+m_Consist.Caption+suf+rsconsist);  end;

procedure TField.mKnow(Sender: TObject);
begin  ShowMessage(pre+m_Know.Caption+suf+rsknow);  end;

procedure TField.mSensors(Sender: TObject);
begin  ShowMessage(pre+m_Sensors.Caption+suf+rssensors);  end;

procedure TField.mFlower(Sender: TObject);
begin  ShowMessage(pre+m_Flower.Caption+suf+rsflower);  end;

procedure TField.mMotion(Sender: TObject);
begin
  AllowMotion:= not AllowMotion;
end;

procedure TField.mExit(Sender: TObject);  begin  Close;  end;
procedure TField.m_ResetListingClick(Sender: TObject); begin  SetDebug(0);  end;

procedure TField.mControl(Sender: TObject);
begin
  m_Motion.Checked:= AllowMotion;
end;

procedure TField.Action;
var  i: integer;
begin
  for  i:= 0  to  ComponentCount - 1   do
  begin
    if  Components[i] is TBall  then  (Components[i] as TBall).Action else
    if  Components[i] is TBall  then  (Components[i] as TBall).Action;
  end;
end;

procedure TField.m_ListingClick(Sender: TObject);
begin
  SetDebug(DebugLevel+1);
end;

procedure TField.SetDebug(level: integer);
begin
  Debuglevel:= level;
  DebugList.Caption:= 'Debuglevel '+ro(Debuglevel);
  if  DebugLevel<=0  then  DebugList.Hide  else  DebugList.Show;
end;

procedure TField.FormDestroy(Sender: TObject);
begin
  TMBody.Destroy;
  BRose.Destroy;
  Butterfly.Destroy;
end;

procedure TField.FormCreate(Sender: TObject);
var  i: integer;
begin
  TMBody:= TMultiBody.Create(Field);{all MultiBody Flyes obey to a common brain}
  for  i:= 1  to  0  do  TMBody.newBody(TEye);
  BRose:= TRose.Create(Field);
  Butterfly:= TButterfly.Create(Field);
end;

end.
